@@ -1,28 +1,22 @@
-use std::io::Result;
+#![feature(allocator_api)]
+
+use std::io::Error;
 use std::os::unix::fs::{FileExt, FileTypeExt};
-use std::os::unix::io::{AsRawFd, RawFd};
+use std::os::unix::io::AsRawFd;
 use std::process::exit;
+use std::vec::Vec;
 
 use libc::{c_ushort, c_int};
+use libc::{sync_file_range, SYNC_FILE_RANGE_WRITE};
 use nix::fcntl::{posix_fadvise, PosixFadviseAdvice};
 use nix::{ioctl_read, ioctl_read_bad, request_code_none};
+use sensitive::alloc::Sensitive;
 
 ioctl_read_bad!(blksectget, request_code_none!(0x12, 103), c_ushort);
 ioctl_read_bad!(blksszget, request_code_none!(0x12, 104), c_int);
 ioctl_read!(blkgetsize64, 0x12, 114, u64);
 
-fn willneed(fd: RawFd, offset: u64, len: u64) -> Result<()> {
-	for advice in [
-		PosixFadviseAdvice::POSIX_FADV_SEQUENTIAL,
-		PosixFadviseAdvice::POSIX_FADV_NOREUSE,
-		PosixFadviseAdvice::POSIX_FADV_WILLNEED] {
-		posix_fadvise(fd, offset.try_into().unwrap(), len.try_into().unwrap(), advice).map(|_| ())?;
-	}
-
-	Ok(())
-}
-
-fn main() -> Result<()> {
+fn main() -> std::io::Result<()> {
 	let mut args = std::env::args();
 
 	if args.len() != 2 {
@@ -93,16 +87,23 @@ fn main() -> Result<()> {
 	assert!(size % ssz as u64 == 0);
 
 	let null = vec![0u8; ssz];
-	let mut buffer = vec![0u8; sect * ssz];
+	let mut buffer = Vec::with_capacity_in(sect * ssz, Sensitive);
 
-	willneed(dev.as_raw_fd(), 0, 0).unwrap_or_else(|err| {
-		eprintln!("Failed to predeclare access pattern for {}: {}", path, err);
-		exit(74);
-	});
+	// The allocator ensures that the memory is zero‐initialised
+	unsafe { buffer.set_len(sect * ssz); }
+
+	for advice in [
+		PosixFadviseAdvice::POSIX_FADV_SEQUENTIAL,
+		PosixFadviseAdvice::POSIX_FADV_WILLNEED] {
+		posix_fadvise(dev.as_raw_fd(), 0, 0, advice).unwrap_or_else(|err| {
+			eprintln!("Failed to predeclare access pattern for {}: {}", path, err);
+			exit(74);
+		});
+	}
 
 	let mut offset = 0u64;
 	let mut verify = 0usize;
-	let mut sync;
+	let mut flush: Option<u64> = None;
 
 	println!();
 
@@ -112,30 +113,28 @@ fn main() -> Result<()> {
 			          bytesize::to_string(offset, true), bytesize::to_string(size, true));
 		}
 
-		// Synchronise to device if last sector of range
-		sync = verify == 1;
-
 		match dev.read_at(if verify == 0 { &mut buffer } else { &mut buffer[0..ssz] }, offset) {
 			Ok(0) => { break; }
 			Ok(len) => {
 				// Assert that we read a multiple of the sector size
 				assert!(len % ssz == 0);
 
+				if verify == 0 {
+					posix_fadvise(dev.as_raw_fd(), offset.try_into().unwrap(), len.try_into().unwrap(),
+					              PosixFadviseAdvice::POSIX_FADV_DONTNEED)
+					.unwrap_or_else(|err| {
+						eprintln!("Failed to declare sector range as no longer used: {}", err);
+						exit(74);
+					});
+				}
+
 				offset += len as u64;
 				verify = verify.saturating_sub(1);
 			}
-
 			Err(err) => {
 				if let Some(libc::EIO) = err.raw_os_error() {
 					if verify == 0 {
 						verify = sect;
-
-						// Declare our intention to re‐read the range
-						willneed(dev.as_raw_fd(), offset, (sect * ssz) as u64).unwrap_or_else(|err| {
-							eprintln!("Failed to predeclare access pattern for range: {}", err);
-							exit(74);
-						});
-
 						continue;
 					}
 
@@ -148,6 +147,11 @@ fn main() -> Result<()> {
 					// Assert that we wrote a complete sector
 					assert_eq!(len, ssz);
 
+					// Remember first sector to flush
+					if flush.is_none() {
+						flush = Some(offset);
+					}
+
 					offset += ssz as u64;
 					verify -= 1;
 				} else {
@@ -157,11 +161,17 @@ fn main() -> Result<()> {
 			}
 		}
 
-		if sync {
-			dev.sync_data().unwrap_or_else(|err| {
-				eprintln!("Failed to synchronise data to device: {}", err);
+		if let Some(start) = flush {
+			if unsafe {
+				sync_file_range(dev.as_raw_fd(), start.try_into().unwrap(),
+				                (offset - start).try_into().unwrap(),
+				                SYNC_FILE_RANGE_WRITE)
+			} != 0 {
+				eprintln!("Failed to flush data to device: {}", Error::last_os_error());
 				exit(74);
-			});
+			}
+
+			flush = None;
 		}
 	}
 
